"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.retryCollectionSubscription = void 0;
const abort_controller_x_1 = require("abort-controller-x");
const isEqual = require("lodash.isequal");
async function* retryCollectionSubscription(subscribe, options = {}) {
    const { signal = new AbortController().signal, baseDelayMs = 1000, maxDelayMs = 30000, maxAttempts = Infinity, onError, getRevision = value => value, equality = isEqual, } = options;
    let state = new Map();
    let attempt = 0;
    let outerCount = -1;
    while (true) {
        outerCount += 1;
        try {
            let innerCount = -1;
            for await (const updates of subscribe(signal)) {
                innerCount += 1;
                attempt = undefined;
                if (innerCount === 0 && outerCount !== 0) {
                    const nextState = new Map();
                    const diff = [];
                    for (const { key, value } of updates) {
                        if (value === undefined) {
                            throw markErrorAsInternal(new Error(`Misbehaving subscription source: unexpected 'undefined' value at key '${key}' in initial emission`));
                        }
                        const revision = getRevision(value);
                        if (!state.has(key) || !equality(state.get(key), revision)) {
                            diff.push({ key, value });
                        }
                        state.delete(key);
                        nextState.set(key, revision);
                    }
                    for (const key of state.keys()) {
                        diff.push({ key });
                    }
                    state = nextState;
                    yield diff;
                }
                else {
                    for (const { key, value } of updates) {
                        if (value === undefined) {
                            if (innerCount === 0) {
                                throw markErrorAsInternal(new Error(`Misbehaving subscription source: unexpected 'undefined' value at key '${key}' in initial emission`));
                            }
                            const deleted = state.delete(key);
                            if (!deleted) {
                                throw markErrorAsInternal(new Error(`Misbehaving subscription source: unexpected 'undefined' value at key '${key}' which was not present in the state`));
                            }
                        }
                        else {
                            state.set(key, getRevision(value));
                        }
                    }
                    yield updates;
                }
            }
            return;
        }
        catch (error) {
            abort_controller_x_1.rethrowAbortError(error);
            if (onError == null && isInternalError(error)) {
                throw error;
            }
            if (attempt === undefined) {
                onError === null || onError === void 0 ? void 0 : onError(error, undefined, undefined);
                attempt = 0;
            }
            else {
                if (attempt >= maxAttempts) {
                    throw error;
                }
                // https://aws.amazon.com/ru/blogs/architecture/exponential-backoff-and-jitter/
                const backoff = Math.min(maxDelayMs, Math.pow(2, attempt) * baseDelayMs);
                const delayMs = Math.round((backoff * (1 + Math.random())) / 2);
                onError === null || onError === void 0 ? void 0 : onError(error, attempt, delayMs);
                await abort_controller_x_1.delay(signal, delayMs);
                attempt += 1;
            }
        }
    }
}
exports.retryCollectionSubscription = retryCollectionSubscription;
const internalErrors = new WeakSet();
function markErrorAsInternal(error) {
    internalErrors.add(error);
    return error;
}
function isInternalError(error) {
    return error instanceof Error && internalErrors.has(error);
}
//# sourceMappingURL=retryCollectionSubscription.js.map